#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2017  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sqlite3
import optparse
import time
import pickle
import re
import pandas as pd
import numpy as np
import os.path
import sys
import helper


class QueryAddIdxes (helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryAddIdxes, self).__init__(dbfname)

    def measure_size(self):
        t = time.time()
        ret = self.c.execute("select count() from reduceDB")
        rows = self.c.fetchall()
        rdb_rows = rows[0][0]
        print("We have %d lines of RDB" % (rdb_rows))

        t = time.time()
        ret = self.c.execute("select count() from clause_stats")
        rows = self.c.fetchall()
        clss_rows = rows[0][0]
        print("We have %d lines of clause_stats" % (clss_rows))

    def create_indexes(self):
        t = time.time()
        print("Recreating indexes...")
        queries = """
        create index `idxclid33` on `sum_cl_use` (`clauseID`, `last_confl_used`);
        ---
        create index `idxclid1` on `clause_stats` (`clauseID`, conflicts, restarts, latest_satzilla_feature_calc);
        create index `idxclid1-2` on `clause_stats` (`clauseID`);
        create index `idxclid1-3` on `clause_stats` (`clauseID`, restarts);
        create index `idxclid1-4` on `clause_stats` (`clauseID`, restarts, prev_restart);
        create index `idxclid1-5` on `clause_stats` (`clauseID`, prev_restart);
        create index `idxclid2` on `clause_stats` (clauseID, `prev_restart`, conflicts, restarts, latest_satzilla_feature_calc);
        create index `idxclid4` on `restart` ( `restarts`);
        create index `idxclid88` on `restart_dat_for_cl` ( `clauseID`);
        create index `idxclid5` on `tags` ( `name`);
        ---
        create index `idxclid6` on `reduceDB` (`clauseID`, conflicts, latest_satzilla_feature_calc);
        create index `idxclid6-2` on `reduceDB` (`clauseID`, `dump_no`);
        create index `idxclid6-3` on `reduceDB` (`clauseID`, `conflicts`, `dump_no`);
        create index `idxclid6-4` on `reduceDB` (`clauseID`, `conflicts`)
        ---
        create index `idxclid7` on `satzilla_features` (`latest_satzilla_feature_calc`);
        ---
        create index `idxclidUCLS-1` on `used_clauses` ( `clauseID`, `used_at`);
        create index `idxclidUCLS-2` on `used_clauses` ( `used_at`);
        ---
        create index `idxcl_last_in_solver-1` on `cl_last_in_solver` ( `clauseID`, `conflicts`);

        """
        for l in queries.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))


class QueryCls (helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryCls, self).__init__(dbfname)
        self.fill_sql_query()

    def fill_sql_query(self):
        # sum_cl_use
        self.sum_cl_use = helper.query_fragment(
            "sum_cl_use", [], "sum_cl_use", options.verbose, self.c)

        # restart data
        not_cols = [
            "simplifications"
            , "restarts"
            , "conflicts"
            , "latest_satzilla_feature_calc"
            , "runtime"
            , "propagations"
            , "decisions"
            , "flipped"
            , "replaced"
            , "eliminated"
            , "set"
            , "free"
            , "numredLits"
            , "numIrredLits"
            , "all_props"
            , "clauseID"
            , "restartID"]
        self.rst_cur = helper.query_fragment(
            "restart_dat_for_cl", not_cols, "rst_cur", options.verbose, self.c)

        # RDB data
        not_cols = [
            "simplifications"
            , "restarts"
            , "conflicts"
            , "latest_satzilla_feature_calc"
            , "runtime"
            , "clauseID"
            , "in_xor"
            , "locked"
            , "activity_rel"]
        self.rdb0_dat = helper.query_fragment(
            "reduceDB", not_cols, "rdb0", options.verbose, self.c)

        # clause data
        not_cols = [
            "simplifications"
            , "restarts"
            , "prev_restart"
            , "latest_satzilla_feature_calc"
            , "antecedents_long_red_age_max"
            , "antecedents_long_red_age_min"
            , "clauseID"]
        self.clause_dat = helper.query_fragment(
            "clause_stats", not_cols, "cl", options.verbose, self.c)

        # satzilla data
        not_cols = [
            "simplifications"
            , "restarts"
            , "conflicts"
            , "latest_satzilla_feature_calc"
            , "irred_glue_distr_mean"
            , "irred_glue_distr_var"]
        self.satzfeat_dat = helper.query_fragment(
            "satzilla_features", not_cols, "szfeat", options.verbose, self.c)

        self.common_limits = """
        order by random()
        limit {limit}
        """

        # final big query
        self.q_select = """
        SELECT
        tags.val as `fname`
        {clause_dat}
        {rst_cur}
        {satzfeat_dat_cur}
        {rdb0_dat}
        {rdb1_dat}
        {sum_cl_use}
        , (rdb0.conflicts - cl.conflicts) as `cl.time_inside_solver`
        , `sum_cl_use`.`last_confl_used`-`cl`.`conflicts` as `x.a_lifetime`
        , used_later_short.used_later_short as `x.used_later_short`
        , used_later_long.used_later_long as `x.used_later_long`
        , used_later_forever.used_later_forever as `x.used_later_forever`
        , sum_cl_use.num_used as `x.sum_cl_use`

        FROM
        reduceDB as rdb0
        -- WARN: ternary clauses are explicity NOT enabled here
        --       since it's a FULL join
        join clause_stats as cl on
            cl.clauseID = rdb0.clauseID

        -- WARN: ternary clauses are explicity NOT enabled here
        --       since it's a FULL join
        join restart_dat_for_cl as rst_cur
            on rst_cur.clauseID = rdb0.clauseID

        join sum_cl_use on
            sum_cl_use.clauseID = rdb0.clauseID

        join reduceDB as rdb1 on
            rdb1.clauseID = rdb0.clauseID

        join satzilla_features as szfeat_cur
            on szfeat_cur.latest_satzilla_feature_calc = cl.latest_satzilla_feature_calc

        join used_later_short on
            used_later_short.clauseID = rdb0.clauseID
            and used_later_short.rdb0conflicts = rdb0.conflicts

        join used_later_long on
            used_later_long.clauseID = rdb0.clauseID
            and used_later_long.rdb0conflicts = rdb0.conflicts

        join used_later_forever on
            used_later_forever.clauseID = rdb0.clauseID
            and used_later_forever.rdb0conflicts = rdb0.conflicts

        join used_later on
            used_later.clauseID = rdb0.clauseID
            and used_later.rdb0conflicts = rdb0.conflicts


        join cl_last_in_solver on
            cl_last_in_solver.clauseID = rdb0.clauseID

        , tags

        WHERE
        cl.clauseID != 0
        and tags.name = "filename"
        and rdb0.dump_no = rdb1.dump_no+1
        and used_later_long.offset = 0
        and used_later_short.offset = 0
        and used_later_forever.offset = 0
        -- and used_later_short_offset.offset = {offset_short}


        -- to avoid missing clauses and their missing data to affect results
        and rdb0.conflicts + {del_at_least} <= cl_last_in_solver.conflicts
        """

        self.myformat = {
            "limit": 1000*1000*1000,
            "clause_dat": self.clause_dat,
            "satzfeat_dat_cur": self.satzfeat_dat.replace("szfeat.", "szfeat_cur."),
            "rdb0_dat": self.rdb0_dat,
            "rdb1_dat": self.rdb0_dat.replace("rdb0", "rdb1"),
            "sum_cl_use": self.sum_cl_use,
            "rst_cur": self.rst_cur,
            "offset_short" : options.short,
            "offset_long" : options.long
        }

    def get_used_later_percentiles(self, name):
        cur = self.conn.cursor()
        q = """
        select
            `type_of_dat`,
            `percentile_descr`,
            `val`
        from used_later_percentiles
        where `type_of_dat` = '{name}'
        """.format(name=name)
        cur.execute(q)
        rows = cur.fetchall()
        lookup = {}
        for row in rows:
            lookup[row[1]] = row[2]
        #print("perc lookup:", lookup)

        return lookup

    def get_one_data_all_dumpnos(self, tier):
        df = None

        ok, df, this_limit = self.get_data(tier)
        if not ok:
            return False, None

        if options.verbose:
            print("Printing head:")
            print(df.head())
            print("Print head done.")

        return True, df


    def get_data(self, tier):
        perc = self.get_used_later_percentiles("_"+tier)

        if tier == "short":
            self.myformat["del_at_least"] = options.short

        elif tier == "long":
            self.myformat["del_at_least"] = options.long

        elif tier == "forever":
            self.myformat["del_at_least"] = 0

        # Make sure these stratas are equally represented
        t = time.time()
        limit = options.limit
        limit, _ = self.run_stratified_queries(limit, perc, tier)
        print("Setting limit to minimum of all of above weighted sizes:", limit)
        print("** Running final queries")
        _, dfs = self.run_stratified_queries(limit, perc, tier)

        data = pd.concat(dfs)
        print("** Queries finished. Total size: %s  -- T: %-3.2f" % (
            data.shape, time.time() - t))
        return True, data , limit

    def run_stratified_queries(self, limit, perc, tier):
        dfs = []
        # NOTE: these are NON-ZERO percentages, but we replace 100 with "0", so the LAST chunk contains ALL, including 0, which is a large part of the data
        for beg_perc, end_perc in [(0, 5), (5, 40), (40, 80), (80, 100)]:
            beg = perc["top_non_zero_{perc}_perc".format(perc=beg_perc)]
            if end_perc == 100:
                end = 0
            else:
                end = perc["top_non_zero_{perc}_perc".format(perc=end_perc)]
            what_to_strata = "`x.used_later_{tier}`".format(tier=tier)

            print("Limit is currently: {limit} value strata perc: ({a}, {b}) translates to value strata ({beg}, {end})".format(
                limit=limit, a=beg_perc, b=end_perc, beg=beg, end=end))

            df, weighted_size = self.query_strata_per_dumpno(
                str(self.q_select),
                limit,
                what_to_strata=what_to_strata,
                strata=(beg,end))
            dfs.append(df)
            # HACK below!
            #limit = int(min(weighted_size))
        return limit, dfs

    def query_strata_per_dumpno(self, q, limit, what_to_strata, strata):
        print("-> Getting one set of data with limit %s" % limit)
        weighted_size = []
        df_parts = []

        def one_part(mult, extra):
            self.myformat["limit"] = int(limit*mult)
            df_parts.append(self.one_query(q + extra, what_to_strata, strata))
            print("--> Num rows for strata %s -- '%s': %s" % (strata, extra, df_parts[-1].shape[0]))

            ws = df_parts[-1].shape[0]/mult
            print("--> The weight was %f so wegthed size is: %d" % (mult, int(ws)))
            weighted_size.append(ws)

        one_part(1/2.0, " and rdb0.dump_no = 1 ")
        one_part(1/4.0, " and rdb0.dump_no = 2 ")
        one_part(1/4.0, " and rdb0.dump_no > 2 ")

        df = pd.concat(df_parts)
        print("-> size of all dump_no-s, strata {strata} data: {size}".format(
            strata=strata, size=df.shape))

        return df, weighted_size

    def one_query(self, q, what_to_strata, strata):
        q = q.format(**self.myformat)
        q = """
        select * from ( {q} )
        where
        {what_to_strata} <= {beg}
        and {what_to_strata} >= {end}""".format(
            q=q,
            what_to_strata=what_to_strata,
            beg=strata[0],
            end=strata[1])

        q += self.common_limits
        q = q.format(**self.myformat)

        t = time.time()
        sys.stdout.write("-> Running query for {} stratas {}...".format(what_to_strata, strata))
        sys.stdout.flush()
        if options.dump_sql:
            print("query:", q)
        df = pd.read_sql_query(q, self.conn)
        print("T: %-3.1f" % (time.time() - t))
        return df


def dump_dataframe(df, name):
    if options.dump_csv:
        fname = "%s.csv" % name
        print("Dumping CSV data to:", fname)
        df.to_csv(fname, index=False, columns=sorted(list(df)))

    fname = "%s.dat" % name
    print("Dumping pandas data to:", fname)
    with open(fname, "wb") as f:
        pickle.dump(df, f)


def one_database(dbfname):
    with QueryAddIdxes(dbfname) as q:
        q.measure_size()
        helper.drop_idxs(q.c)
        q.create_indexes()

    with helper.QueryFill(dbfname) as q:
        q.delete_and_create_all()
        q.fill_used_later()
        q.fill_used_later_X("long", offset=0, duration=options.long)
        q.fill_used_later_X("short", offset=0, duration=options.short)
        q.fill_used_later_X("forever", offset=0, duration=(1000*1000*1000), forever=True)

        q.fill_used_later_X("long", offset=options.long, duration=options.long)
        q.fill_used_later_X("short", offset=options.short, duration=options.short)

    print("Using sqlite3 DB file %s" % dbfname)
    for tier in ["short", "long", "forever"]:
        print("------> Doing tier {tier}".format(tier=tier))

        with QueryCls(dbfname) as q:
            ok, df = q.get_one_data_all_dumpnos(tier)

        if not ok:
            print("-> Skipping file {file} {tier}".format(
                file=dbfname, tier=tier))
            continue

        if options.verbose:
            print("Describing----")
            dat = df.describe()
            print(dat)
            print("Describe done.---")
            print("Features: ", df.columns.values.flatten().tolist())

        if options.verbose:
            print("Describing post-transform ----")
            print(df.describe())
            print("Describe done.---")
            print("Features: ", df.columns.values.flatten().tolist())

        cleanname = re.sub(r'\.cnf.gz.sqlite$', '', dbfname)
        cleanname = re.sub(r'\.db$', '', dbfname)
        cleanname = re.sub(r'\.sqlitedb$', '', dbfname)
        cleanname = "{cleanname}-cldata-{tier}".format(
            cleanname=cleanname,
            tier=tier)

        dump_dataframe(df, cleanname)


if __name__ == "__main__":
    usage = "usage: %prog [options] file1.sqlite [file2.sqlite ...]"
    parser = optparse.OptionParser(usage=usage)

    # verbosity level
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--sql", action="store_true", default=False,
                      dest="dump_sql", help="Dump SQL queries")
    parser.add_option("--csv", action="store_true", default=False,
                      dest="dump_csv", help="Dump CSV (for weka)")

    # limits
    parser.add_option("--limit", default=20000, type=int,
                      dest="limit", help="Exact number of examples to take. -1 is to take all. Default: %default")

    # debugging is faster with this
    parser.add_option("--noind", action="store_true", default=False,
                      dest="no_recreate_indexes",
                      help="Don't recreate indexes")

    # lengths of short/long
    parser.add_option("--short", default="10000", type=str,
                      dest="short", help="Short duration. Default: %default")
    parser.add_option("--long", default="50000", type=str,
                      dest="long", help="Long duration. Default: %default")

    (options, args) = parser.parse_args()

    if len(args) != 1:
        print("ERROR: You must give exactly one file")
        exit(-1)

    np.random.seed(2097483)
    one_database(args[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import optparse
import random
import time
import string
import configparser


def parse_arguments():
    class PlainHelpFormatter(optparse.IndentedHelpFormatter):

        def format_description(self, description):
            if description:
                return description + "\n"
            else:
                return ""

    usage = """usage: %prog

To use other solvers, give:
--solver SWDiA5BY.alt.vd.res.va2.15000.looseres.3tierC5/binary/SWDiA5BY_static.alt.vd
--solver SWDiA5BY_A26/binary/SWDiA5BY_static_A26
--solver lingeling_ayv/binary/lingeling_ayv
--solver glucose2016/simp/glucose_static_2016
--solver MapleCOMSPS/simp/maplecomsps_static
--solver cmsat-satcomp16/bin/cryptominisat4_simple
--solver lingeling-bbc/build/lingeling/lingeling_bbc
--solver Maple_LCM_Dist/Maple_LCM_Dist # may work with --drat, but needs updated DRAT checker


Use-cases:
# normal run
./launch_server.py --cnflist satcomp17_updated

# stats run
./launch_server.py --cnf test_updated --stats --drat --tout 600 --memlimit 10000
./launch_server.py --cnf unsat_small_candidates_fullpath --stats --drat --tout 600 --memlimit 10000


# testing, using small instance to check (cheaper & faster)
./launch_server.py --cnflist test_updated

# 2 clients, no preprocessing
./launch_server.py --cnflist satcomp14 -c 2 --opt "--preproc 0" --folder no_preproc

# gaussian elimination -- automatic detection, built with GAUSS
./launch_server.py --cnflist satcomp14 --folder gauss

# clause IDs so learning can be performed -- gzipped SQL output with clause IDs will be produced
./launch_server.py --stats --drat --folder learning

# to give options to the solver
./launch_server.py --folder with_opts --opt \"--ml=1,--keepglue=4\""

 # to upload satzilla_features_to_reconf.cpp
aws s3 cp ../../src/satzilla_features_to_reconf.cpp s3://msoos-solve-data/solvers/

"""
    parser = optparse.OptionParser(usage=usage, formatter=PlainHelpFormatter())
    parser.add_option("--verbose", "-v", action="store_true",
                      default=False, dest="verbose", help="Be more verbose"
                      )

    parser.add_option("--numclients", "-c", default=None, type=int,
                      dest="client_count", help="Number of clients to launch"
                      )

    parser.add_option("--port", "-p", default=10000, dest="port",
                      help="Port to listen on. [default: %default]", type="int"
                      )

    parser.add_option("--tout", "-t", default=3000, dest="timeout_in_secs",
                      help="Timeout for the file in seconds"
                      "[default: %default]",
                      type=int
                      )

    parser.add_option("--toutmult", default=12.1, dest="tout_mult",
                      help="Approx: 1x is solving, 10x time is DRAT time wait, 1x is parsing, 0.1x that is sending us the result."
                      "[default: %default]",
                      type=float
                      )

    parser.add_option("--memlimit", "-m", default=1600, dest="mem_limit_in_mb",
                      help="Memory limit in MB"
                      "[default: %default]",
                      type=int
                      )

    parser.add_option("--cnflist", default="satcomp14_updated", dest="cnf_list",
                      type=str,
                      help="The list of CNF files to solve, first line the dir"
                      "[default: %default]",
                      )

    parser.add_option("--dir", default="/home/ubuntu/", dest="base_dir", type=str,
                      help="The home dir of cryptominisat [default: %default]"
                      )

    parser.add_option("--solver",
                      default="cryptominisat/build/cryptominisat5",
                      dest="solver",
                      help="Solver executable"
                      "[default: %default]",
                      type=str
                      )

    parser.add_option("--folder", default="results", dest="given_folder",
                      help="S3 folder name to upload data"
                      "[default: %default]",
                      type=str
                      )

    parser.add_option("--git", dest="git_rev", type=str,
                      help="The GIT revision to use. Default: HEAD"
                      )

    parser.add_option("--opt", dest="extra_opts", type=str, default="",
                      help="Extra options to give to solver"
                      "[default: %default]",
                      )

    parser.add_option("--noshutdown", "-n", default=False, dest="noshutdown",
                      action="store_true", help="Do not shut down clients"
                      )

    parser.add_option("--drat", default=False, dest="drat",
                      action="store_true", help="Use DRAT"
                      )

    parser.add_option("--stats", default=False, dest="stats",
                      action="store_true", help="Use STATS and get SQLITE data"
                      )

    parser.add_option("--gauss", default=False, dest="gauss",
                      action="store_true", help="Use GAUSS"
                      )

    parser.add_option("--logfile", dest="logfile_name", type=str,
                      default="python_server_log.log", help="Name of LOG file")

    # parse options
    options, args = parser.parse_args()
    conf = configparser.ConfigParser()
    if options.cnf_list == "test":
        conf.read('ec2-spot-instance-test.cfg')
    else:
        conf.read('ec2-spot-instance.cfg')

    options.s3_bucket = conf.get("ec2", "result_bucket")
    options.key_name = conf.get("ec2", "key_name")
    options.security_group_server = conf.get("ec2", "security_group_server")
    options.subnet_id = conf.get("ec2", "subnet_id")
    options.ami_id = conf.get("ec2", "ami_id")
    options.region = conf.get("ec2", "region")

    def rnd_id():
        return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5))

    options.logfile_name = options.base_dir + options.logfile_name
    options.given_folder += "-" + time.strftime("%d-%B-%Y")
    options.given_folder += "-%s" % rnd_id()
    options.given_folder += "-%s" % options.cnf_list

    if options.drat and not options.stats:
        print("ERROR: You must have --stats when you use --drat")
        exit(-1)

    return options, args


if __name__ == "__main__":
    options, args = parse_arguments()
    print("Options are:", options)
    print("args are:", args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2017  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sqlite3
import argparse
import os
import struct
import time


class Query:
    def __init__(self, dbfname):
        self.conn = sqlite3.connect(dbfname)
        self.c = self.conn.cursor()

        self.cl_used = []
        self.cl_used_num = 0
        self.cl_used_total = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.commit()
        self.conn.close()

    def delete_tbls(self):
        queries = """
        delete from used_clauses;
        """
        for l in queries.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Runnin query: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Query T: %-3.2f s" % (time.time() - t2))

    def get_last_good(self, basefname):
        last_good = -1
        for i in range(100):
            tfname = "%s-%d" % (basefname, i)
            print("Checking if file %s exists" % tfname)
            if not os.path.isfile(tfname):
                break
            last_good = i

        if last_good == -1:
            print("ERROR: file does not exist at all")
            exit(-1)

        return last_good

    def add_used_clauses(self, usedClFname):
        last_good = self.get_last_good(usedClFname)
        tfname = "%s-%d" % (usedClFname, last_good)
        print("Adding data from used_clauses file ", tfname)
        t = time.time()

        self.cl_used = []
        self.cl_used_num = 0
        self.cl_used_total = 0
        with open(tfname, "rb") as f:
            while True:
                b1 = f.read(8)
                if not b1:
                    break
                b2 = f.read(8)
                cl_id = struct.unpack("<q", b1)[0]
                conf = struct.unpack("<q", b2)[0]
                self.cl_used.append((cl_id, conf))
                self.cl_used_num += 1
                self.cl_used_total += 1
                if self.cl_used_num > 10000:
                    self.dump_used_clauses()

        self.dump_used_clauses()
        print("Added use data: %d time: T: %-3.2f s" % (self.cl_used_total, time.time() - t))

    def dump_used_clauses(self):
        self.c.executemany("""
        INSERT INTO used_clauses (
        `clauseID`
        , `used_at`)
        VALUES (?, ?);""", self.cl_used)
        self.cl_used = []
        self.cl_used_num = 0


if __name__ == "__main__":
    usage = """usage: %(prog)s [options] sqlite_db usedCls

Adds used_clauses to the SQLite database"""

    parser = argparse.ArgumentParser(usage=usage)


    parser.add_argument("sqlitedb", type=str, metavar='USEDCLS')
    parser.add_argument("usedcls", type=str, metavar='USEDCLS')
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    options = parser.parse_args()

    if options.sqlitedb is None:
        print("Error you must give an SQLite DB. Please follow usage.")
        print(usage)
        exit(-1)

    if options.usedcls is None:
        print("Error you must give an Used CLS list. Please follow usage.")
        print(usage)
        exit(-1)

    print("Using sqlite3db file %s" % options.sqlitedb)
    print("Base used_clauses file is %s" % options.usedcls)

    with Query(options.sqlitedb) as q:
        q.delete_tbls()
        q.add_used_clauses(options.usedcls)

    print("Finished adding good lemma indicators to db %s" % options.sqlitedb)

    exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sys
import boto.ec2
import os
import subprocess
import server_option_parser


def get_answer():
    yes = set(['yes', 'y', 'ye'])
    no = set(['no', 'n'])

    choice = input().lower()
    if choice in yes:
        return True
    elif choice in no:
        return False
    else:
        sys.stdout.write("Please respond with 'yes' or 'no'\n")
        exit(0)


def push():
    print("First we push, oherwise we'll forget...")
    ret = os.system("git push")
    if ret != 0:
        print("Oops, couldn't push, exiting before executing")
        exit(-1)

    print("")


if __name__ == "__main__":
    options, args = server_option_parser.parse_arguments()
    print("Options are:")
    for a, b in options.__dict__.items():
        print("-- %-30s : %s" % (a, b))
    assert args == []

    if options.mem_limit_in_mb < 10000 and options.drat:
        print("******* WARNING ********")
        print("Beware: your memory is WAY too low for DRAT and learning stuff")

    push()
    data = ""
    opt_is_on = False
    for x in sys.argv[1:]:
        if opt_is_on:
            data += " ".join(x.split(","))
            data += "\" "
            opt_is_on = False
            continue
        if x != "--opt":
            data += x + " "
            continue

        opt_is_on = True
        data += "--opt \""

    if ("--git" not in data):
        revision = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode("utf-8")
        data += " --git '{revision}'".format(revision=revision)

    if len(sys.argv) > 1:
        print("Launching with data: %s" % data)
    else:
        print("you must give at least one parameter, probably --s3folder")
        exit(-1)

    sys.stdout.write("Is this OK? [y/n]? ")
    if not get_answer():
        print("Aborting")
        exit(0)

    print("Executing!")

    cloud_init = """#!/bin/bash
set -e

apt-get update
apt-get install -y python
apt-get -y install git python-pip
pip install --force-reinstall --upgrade awscli
pip install --force-reinstall --upgrade boto
pip install configparser

# Get AWS log agent
cd /home/ubuntu/

cat > aws-logs-server.conf << EOF
[general]
state_file = /home/ubuntu/cloudwatch.state

[logstream1]
log_group_name = cyrptominisat-perftest
log_stream_name = server
file = /home/ubuntu/*.log

[messages]
log_group_name = cyrptominisat-perftest
log_stream_name = server
file = /var/log/messages
EOF

curl https://s3.amazonaws.com/aws-cloudwatch/downloads/latest/awslogs-agent-setup.py -O
python ./awslogs-agent-setup.py --region {region} -c aws-logs-server.conf -n

sudo -H -u ubuntu bash -c 'ssh-keyscan github.com >> ~/.ssh/known_hosts'
sudo -H -u ubuntu bash -c 'git clone --depth 50 https://github.com/msoos/cryptominisat.git'

# Get credentials
cd /home/ubuntu
sudo -H -u ubuntu bash -c 'aws s3 cp s3://msoos-solve-data/solvers/email.conf . --region={region}'

# Start server
cd /home/ubuntu/cryptominisat
sudo -H -u ubuntu bash -c '/home/ubuntu/cryptominisat/scripts/aws/pre-server.py > /home/ubuntu/pre_server_log.log  2>&1 &'

DATA="{data}"
    """.format(region=options.region, data=data)

    conn = boto.ec2.connect_to_region(options.region)
    conn.run_instances(
        min_count=1,
        max_count=1,
        image_id=options.ami_id,
        subnet_id=options.subnet_id,
        instance_type='t2.micro',
        instance_profile_arn='arn:aws:iam::907572138573:instance-profile/server',
        user_data=cloud_init,
        key_name=options.key_name,
        security_group_ids=[options.security_group_server],
        instance_initiated_shutdown_behavior='terminate')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# pylint: disable=invalid-name,line-too-long,too-many-locals,consider-using-sys-exit

import operator
import re
import time
import argparse
import sys
import os
import itertools
import pandas as pd
import pickle
import sklearn
import sklearn.svm
import sklearn.tree
from sklearn.preprocessing import StandardScaler
import numpy as np
import sklearn.metrics
import matplotlib.pyplot as plt
import sklearn.ensemble
import sklearn.linear_model
import helper
import xgboost as xgb
import ast
import functools
import crystalcodegen as ccg
try:
    import mlflow
except ImportError:
    mlflow_avail = False
else:
    mlflow_avail = True

ver = sklearn.__version__.split(".")
if int(ver[1]) < 20:
    from sklearn.cross_validation import train_test_split
else:
    from sklearn.model_selection import train_test_split

MISSING=np.nan

def check_long_short():
    if options.tier is None:
        print("ERROR: You must give option '--tier' as short/long/forever")
        assert False
        exit(-1)


class Learner:
    def __init__(self, df, tier):
        self.df = df
        self.tier = tier

    def count_bad_ok(self, df):
        files = df[["x.class", "rdb0.dump_no"]].groupby("x.class").count()
        if files["rdb0.dump_no"].index[0] == 0:
            bad = files["rdb0.dump_no"][0]
            ok = files["rdb0.dump_no"][1]
        else:
            bad = files["rdb0.dump_no"][1]
            ok = files["rdb0.dump_no"][0]

        assert bad > 0, "No need to train, data only contains BAD(0)"
        assert ok > 0, "No need to train, data only contains OK(1)"

        return bad, ok

    def filter_percentile(self, df, features, perc):
        low = df.quantile(perc, axis=0)
        high = df.quantile(1.0-perc, axis=0)
        df2 = df.copy()
        for i in features:
            df2 = df2[(df2[i] >= low[i]) & (df2[i] <= high[i])]
            print("Filtered to %f on %-30s, shape now: %s" %
                  (perc, i, df2.shape))

        print("Original size:", df.shape)
        print("New size:", df2.shape)
        return df2

    def filtered_conf_matrixes(self, dump_no, data, features, to_predict, clf,
                               toprint, highlight=False):
        # filter test data
        if dump_no is not None:
            print("\nCalculating confusion matrix -- dump_no == %s" % dump_no)
            toprint += " dump no %d" % dump_no
            data2 = data[data["rdb0.dump_no"] == dump_no]
        else:
            print("\nCalculating confusion matrix -- ALL dump_no")
            data2 = data

        if False:
            return helper.conf_matrixes(data2, features, to_predict, clf, toprint,
                                    highlight=highlight)
        else:
            return helper.calc_regression_error(data2, features, to_predict, clf, toprint,
                                    highlight=highlight)

    @staticmethod
    def fix_feat_name(x):
        x = re.sub(r"dump_no", r"dump_no", x)
        x = re.sub(r"^cl_", r"", x)
        x = re.sub(r"^rdb0_", r"", x)

        if x == "dump_no":
            pass
        elif x == "last_touched_diff":
            pass
        elif x == "act_ranking_top_10":
            pass
        elif x == "act_ranking_rel":
            pass
        elif x == "time_inside_solver":
            pass
        elif x == "size":
            x = "cl->" + x + "()"
        else:
            x = "cl->stats." + x

        return x

    def importance_XGB(self, clf, features):
        impdf = []
        #print("clf:", clf)
        #print("clf-booster:", clf.feature_importances_)

        for i in range(len(clf.feature_importances_)):
            score = clf.feature_importances_[i]
            ft = features[i]
            impdf.append({'feature': ft, 'importance': score})
        impdf = pd.DataFrame(impdf)
        impdf = impdf.sort_values(by='importance', ascending=False).reset_index(drop=True)
        impdf['importance'] /= impdf['importance'].sum()
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)
        pd.set_option('display.width', None)
        pd.set_option('display.max_colwidth', None)
        print("impdf:", impdf)

        return impdf

    def one_classifier(self, features, to_predict, final):
        print("-> Number of features  :", len(features))
        print("-> Number of datapoints:", self.df.shape)
        print("-> Predicting          :", to_predict)
        df = self.df.copy()

        if options.check_row_data:
            helper.check_too_large_or_nan_values(df, features)

        # count bad and ok
        #bad,ok = self.count_bad_ok(df)
        #print("Number of BAD  elements        : %-6d" % bad)
        #print("Number of OK   elements        : %-6d" % ok)
        # balance it out
        #prefer_ok = float(bad)/float(ok)
        #print("Balanced OK preference would be: %-6.3f" % prefer_ok)
        # apply inbalance from option given
        #prefer_ok *= options.prefer_ok
        #print("Option to prefer OK is set to  : %-6.3f" % options.prefer_ok)
        #print("Final OK preference is         : %-6.3f" % prefer_ok)

        print("Value distribution of 'dump_no':\n%s" % df["rdb0.dump_no"].value_counts())
        print("Value distribution of to_predict:\n%s" % df[to_predict].value_counts())
        train, test = train_test_split(df, test_size=0.33, random_state=prng)

        X_train = train[features]
        y_train = train[to_predict]

        t = time.time()
        clf = None

        if final:
            split_point = helper.calc_min_split_point(
                df, options.min_samples_split)

            clf_tree = sklearn.tree.DecisionTreeClassifier(
                max_depth=options.tree_depth,
                min_samples_split=split_point,
                random_state=prng)
            clf_svm_pre = sklearn.svm.SVC(
                C=500,
                gamma=10**-5,
                random_state=prng)
            clf_svm = sklearn.ensemble.BaggingClassifier(
                clf_svm_pre,
                n_estimators=3,
                max_samples=0.5, max_features=0.5,
                random_state=prng)
            clf_logreg = sklearn.linear_model.LogisticRegression(
                C=0.3,
                random_state=prng)
            clf_forest = sklearn.ensemble.RandomForestRegressor(
                n_estimators=options.num_trees,
                max_features="sqrt",
                #min_samples_leaf=split_point,
                random_state=prng)

            if options.final_is_tree:
                clf = clf_tree
            elif options.final_is_svm:
                clf = clf_svm
            elif options.final_is_logreg:
                clf = clf_logreg
            elif options.final_is_forest:
                clf = clf_forest
            elif options.final_is_xgboost:
                clf = xgb.XGBRegressor(objective='reg:squarederror', missing=MISSING)
            elif options.final_is_voting:
                mylist = [["forest", clf_forest], [
                    "svm", clf_svm], ["logreg", clf_logreg]]
                clf = sklearn.ensemble.VotingClassifier(mylist, random_state=prng)
            else:
                print(
                    "ERROR: You MUST give one of: tree/forest/svm/logreg/voting classifier")
                exit(-1)
        else:
            assert options.final_is_forest or options.final_is_xgboost, "For TOP calculation, we must have --forest or --xgboost"
            clf_forest = sklearn.ensemble.RandomForestRegressor(
                n_estimators=options.num_trees*5,
                max_features="sqrt",
                random_state=prng)

            clf = clf_forest

        del df
        #my_sample_w = np.array(y_train.values)
        #my_sample_w = my_sample_w*prefer_ok
        #def zero_to_one(t):
            #if t == 0:
                #return 1
            #else:
                #return t
        #vfunc = np.vectorize(zero_to_one)
        #my_sample_w = vfunc(my_sample_w)
        #clf.fit(X_train, y_train, sample_weight=my_sample_w)
        clf.fit(X_train, y_train)

        print("Training finished. T: %-3.2f" % (time.time() - t))

        if not final:
            if mlflow_avail:
                mlflow.log_param("features used", features)
                # mlflow.log_metric("all features: ", train.columns.values.flatten().tolist())
                mlflow.log_metric("train num rows", train.shape[0])
                mlflow.log_metric("test num rows", test.shape[0])

            #mlflow.log_artifact("best_features", best_features)
        else:
            if options.dot is not None:
                if not options.final_is_tree:
                    print("ERROR: You cannot use the DOT function on non-trees")
                    exit(-1)

                helper.output_to_classical_dot(
                    clf, features,
                    fname=options.dot + "-" + self.func_name)

            if options.basedir and options.final_is_xgboost:
                booster = clf.get_booster()
                fname = options.basedir + "/predictor_{name}.json".format(
                    name=options.name)
                booster.save_model(fname)
                print("==> Saved model to: ", fname)
            else:
                print("WARNING: NOT writing code -- you must use xgboost and give dir for that")


        # print feature rankings
        if options.final_is_forest:
            helper.print_feature_ranking(
                clf, X_train,
                top_num_features=200,
                features=features,
                plot=options.show)
        elif options.final_is_xgboost:
            self.importance_XGB(clf, features=features)

        # print distribution of error
        print("--------------------------")
        print("-       test data        -")
        print("--------------------------")
        for dump_no in [1, 2, 3, 10, 20, 40, None]:
            self.filtered_conf_matrixes(
                dump_no, test, features, to_predict, clf, "test data", highlight=True)
        print("--------------------------------")
        print("--      train+test data        -")
        print("--------------------------------")
        for dump_no in [1, None]:
            self.filtered_conf_matrixes(
                dump_no, self.df, features, to_predict, clf, "test and train data")
        print("--------------------------")
        print("-       train data       -")
        print("--------------------------")
        self.filtered_conf_matrixes(
            dump_no, train, features, to_predict, clf, "train data")

        print("Example data:")
        #print(test[features+to_predict])
        #print(test[to_predict])
        with open("example-data.dat", "wb") as f:
            pickle.dump(test[features+[to_predict]], f)


        # Calculate predicted value for the original dataframe
        y_pred = clf.predict(self.df[features])
        return y_pred

    def rem_features(self, feat, to_remove):
        feat_less = list(feat)
        for feature in feat:
            for rem in to_remove:
                if rem in feature:
                    feat_less.remove(feature)
                    if options.verbose:
                        print("Removing feature from feat_less:", feature)

        return feat_less

    def learn(self):
        if options.raw_data_plots:
            pd.options.display.mpl_style = "default"
            self.df.hist()
            self.df.boxplot()

        if not options.only_final:
            features = list(self.df)
            features = self.rem_features(
            features, ["x.class",
                       "x.a_lifetime",
                       "fname",
                       "sum_cl_use.num_used",
                       "x.sum_cl_use",
                       "x.used_later_short",
                       "x.used_later_long",
                       "x.used_later_forever"])

            self.one_classifier(
                features,
                "x.used_later_{name}".format(name=options.tier),
                final=False)

        else:
            best_features = helper.get_features(options.best_features_fname)
            self.one_classifier(
                best_features,
                "x.used_later_{name}".format(name=options.tier),
                final=True)

            if options.show:
                plt.show()


if __name__ == "__main__":
    usage = "usage: %(prog)s [options] file.pandas"
    parser = argparse.ArgumentParser(usage=usage)

    parser.add_argument("fname", type=str, metavar='PANDASFILE')
    parser.add_argument("--seed", default=None, type=int,
                        dest="seed", help="Seed of PRNG")
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                        dest="verbose", help="Print more output")
    parser.add_argument("--printfeat", action="store_true", default=False,
                        dest="print_features", help="Print features")
    parser.add_argument("--nocomputed", default=False, action="store_true",
                        dest="no_computed", help="Don't add computed features")
    parser.add_argument("--allcomputed", default=False, action="store_true",
                        dest="all_computed", help="Add ALL computed features")
    parser.add_argument("--bestfeatfile", default="../../scripts/crystal/best_features.txt", type=str,
                        dest="best_features_fname", help="Name and position of best features file that lists the best features in order")

    # tree/forest options
    parser.add_argument("--depth", default=None, type=int,
                        dest="tree_depth", help="Depth of the tree to create")
    parser.add_argument("--split", default=0.01, type=float, metavar="RATIO",
                        dest="min_samples_split", help="Split in tree if this many samples or above. Used as a percentage of datapoints")
    parser.add_argument("--numtrees", default=100, type=int,
                        dest="num_trees", help="How many trees to generate for the forest")

    # generation of predictor
    parser.add_argument("--dot", type=str, default=None,
                        dest="dot", help="Create DOT file")
    parser.add_argument("--filterdot", default=0.05, type=float,
                        dest="filter_dot", help="Filter the DOT output from outliers so the graph looks nicer")
    parser.add_argument("--show", action="store_true", default=False,
                        dest="show", help="Show visual graphs")
    parser.add_argument("--check", action="store_true", default=False,
                        dest="check_row_data", help="Check row data for NaN or float overflow")
    parser.add_argument("--rawplots", action="store_true", default=False,
                        dest="raw_data_plots", help="Display raw data plots")
    parser.add_argument("--basedir", type=str,
                        dest="basedir", help="The base directory of where the CryptoMiniSat source code is")

    # data filtering
    parser.add_argument("--only", default=1.00, type=float,
                        dest="only_perc", help="Only use this percentage of data")

    # final generator top/final
    parser.add_argument("--final", default=False, action="store_true",
                        dest="only_final", help="Only generate final predictor")
    parser.add_argument("--top", default=None, type=int, metavar="TOPN",
                        dest="top_num_features", help="Candidates are top N features for greedy selector")

    # type of classifier
    parser.add_argument("--tree", default=False, action="store_true",
                        dest="final_is_tree", help="Final classifier should be a tree")
    parser.add_argument("--svm", default=False, action="store_true",
                        dest="final_is_svm", help="Final classifier should be an svm")
    parser.add_argument("--logreg", default=False, action="store_true",
                        dest="final_is_logreg", help="Final classifier should be a logistic regression")
    parser.add_argument("--forest", default=False, action="store_true",
                        dest="final_is_forest", help="Final classifier should be a forest")
    parser.add_argument("--xgboost", default=False, action="store_true",
                        dest="final_is_xgboost", help="Final classifier should be a XGBoost")
    parser.add_argument("--voting", default=False, action="store_true",
                        dest="final_is_voting", help="Final classifier should be a voting of all of: forest, svm, logreg")

    # which one to generate
    parser.add_argument("--tier", default=None, type=str,
                        dest="tier", help="what to predict")
    parser.add_argument("--name", default=None, type=str,
                        dest="name", help="what file to generate")

    options = parser.parse_args()
    prng = np.random.RandomState(options.seed)

    if options.fname is None:
        print("ERROR: You must give the pandas file!")
        exit(-1)

    if options.top_num_features and options.only_final:
        print("Can't do both --top and --final")
        exit(-1)

    if options.top_num_features is None and not options.only_final:
        print("You must have either --top features OR --final")
        exit(-1)

    assert options.min_samples_split <= 1.0, "You must give min_samples_split that's smaller than 1.0"
    if not os.path.isfile(options.fname):
        print("ERROR: '%s' is not a file" % options.fname)
        exit(-1)

    if options.name is None:
        print("Name was not set with --name, setting it to --tier, i.e. ", options.tier)
        options.name = options.tier

    # ------------
    #  Log all parameters
    # ------------
    if mlflow_avail:
        mlflow.log_param("final", options.only_final)
        if options.only_final:
            mlflow.log_param("tier", options.tier)
            mlflow.log_param("name", options.name)
            mlflow.log_param("tree", options.final_is_tree)
            mlflow.log_param("svm", options.final_is_svm)
            mlflow.log_param("logreg", options.final_is_logreg)
            mlflow.log_param("forest", options.final_is_forest)
            mlflow.log_param("voting", options.final_is_voting)
            mlflow.log_param("basedir", options.basedir)
        else:
            mlflow.log_param("top_num_features", options.top_num_features)

        #mlflow.log_param("prefer_ok", options.prefer_ok)
        mlflow.log_param("only_percentage", options.only_perc)
        mlflow.log_param("min_samples_split", options.min_samples_split)
        mlflow.log_param("tree_depth", options.tree_depth)
        mlflow.log_param("num_trees", options.num_trees)
        mlflow.log_param("no_computed", options.no_computed)
        mlflow.log_artifact(options.fname)

    # Read in Pandas Dataframe
    print("Reading dataframe...")
    df = pd.read_pickle(options.fname)
    df_orig = df.copy()
    if options.print_features:
        for f in sorted(list(df)):
            print(f)

    # get smaller part to work on
    # also, copy it so we don't get warning about setting a slice of a DF
    print("Applying only...")
    df_tmp = df.sample(frac=options.only_perc, random_state=prng)
    df = pd.DataFrame(df_tmp)
    print("-> Number of datapoints after applying '--only':", df.shape)

    # feature manipulation
    if not options.no_computed:
        if options.all_computed:
            helper.cldata_add_computed_features(df, options.verbose)
        else:
            print("Adding features...")
            divide = functools.partial(helper.helper_divide, df=df, features=list(df), verb=options.verbose)
            rdb0_act_ranking_rel = divide("rdb0.act_ranking", "rdb0.tot_cls_in_db", name="rdb0_act_ranking_rel")
            rdb1_act_ranking_rel = divide("rdb1.act_ranking", "rdb1.tot_cls_in_db", name="rdb1_act_ranking_rel")
            best_features = helper.get_features(options.best_features_fname)
            for feat in best_features:
                toeval = ccg.to_source(ast.parse(feat))
                print("Adding feature %s as eval %s" % (feat, toeval))
                df[feat] = eval(toeval)

    helper.clear_data_from_str(df)

    print("Filling NA with MISSING..")
    df.fillna(MISSING, inplace=True)
    df.replace(np.NINF, MISSING, inplace=True)
    df.replace(np.NaN, MISSING, inplace=True)
    df.replace(np.Infinity, MISSING, inplace=True)
    df.replace(np.inf, MISSING, inplace=True)



    # do the heavy lifting
    learner = Learner(df, tier=options.tier)
    learner.learn()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import boto.utils
import os

user_data = boto.utils.get_instance_userdata()

todo = ""
for line in user_data.split("\n"):
    if "DATA" in line:
        todo = line[5:].strip().strip('"')

if todo == "":
    exit(0)

os.chdir("/home/ubuntu/cryptominisat/scripts/aws/")
command = "nohup /home/ubuntu/cryptominisat/scripts/aws/server.py %s &" % todo
os.system(command)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import os
import socket
import sys
import struct
import pickle
import time
import pprint
import traceback
import Queue
import threading
import logging
import server_option_parser

# for importing in systems where "." is not in the PATH
sys.path.append(os.getcwd())
from common_aws import *
import RequestSpotClient


def get_n_bytes_from_connection(sock, MSGLEN):
    chunks = []
    bytes_recd = 0
    while bytes_recd < MSGLEN:
        chunk = sock.recv(min(MSGLEN - bytes_recd, 2048))
        if chunk == '':
            raise RuntimeError("socket connection broken")
        chunks.append(chunk)
        bytes_recd = bytes_recd + len(chunk)

    return ''.join(chunks)


def send_command(sock, command, tosend=None):
    # note, this is a python issue, we can't set above tosend={}
    # https://nedbatchelder.com/blog/200806/pylint.html
    tosend = tosend or {}

    tosend["command"] = command
    tosend = pickle.dumps(tosend)
    tosend = struct.pack('!q', len(tosend)) + tosend
    sock.sendall(tosend)


class ToSolve:

    def __init__(self, num, name):
        self.num = num
        self.name = name

    def __str__(self):
        return "%s (num: %d)" % (self.name, self.num)


class Server (threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.files_available = []
        self.files_finished = []
        self.files = {}

        logging.info("Getting list of files %s", options.cnf_list)
        key = boto.connect_s3().get_bucket("msoos-solve-data").get_key("solvers/" + options.cnf_list)
        key.get_contents_to_filename(options.cnf_list)

        fnames = open(options.cnf_list, "r")
        logging.info("CNF list is file %s", options.cnf_list)
        num = 0
        for fname in fnames:
            fname = fname.strip()
            self.files[num] = ToSolve(num, fname)
            self.files_available.append(num)
            logging.info("File added: %s", fname)
            num = num+1
        fnames.close()

        self.files_running = {}
        logging.info("Solving %d files", len(self.files_available))
        self.uniq_cnt = 0

    def ready_to_shutdown(self):
        if len(self.files_available) > 0:
            return False

        if len(self.files_finished) < len(self.files):
            return False

        return True

    def handle_done(self, connection, cli_addr, indata):
        file_num = indata["file_num"]

        logging.info("Finished with file %s (num %d), got files %s",
                     self.files[indata["file_num"]], indata["file_num"],
                     indata["files"])
        self.files_finished.append(indata["file_num"])
        if file_num in self.files_running:
            del self.files_running[file_num]

        logging.info("Num files_available: %d Num files_finished %d",
                     len(self.files_available), len(self.files_finished))

        self.rename_files_to_final(indata["files"])
        sys.stdout.flush()

    def rename_files_to_final(self, files):
        for fnames in files:
            logging.info("Renaming file %s to %s", fnames[0], fnames[1])
            ret = os.system("aws s3 mv s3://{bucket}/{origname} s3://{bucket}/{toname} --region {region}".format(
                bucket=options.s3_bucket,
                origname=fnames[0],
                toname=fnames[1],
                region=options.region))
            if ret:
                logging.warn("Renaming file to final name failed!")

    def check_for_dead_files(self):
        this_time = time.time()
        files_to_remove_from_files_running = []
        for file_num, starttime in self.files_running.items():
            duration = this_time - starttime
            # print("* death check. running:" , file_num, " duration: ",
            # duration)
            if duration > options.timeout_in_secs*options.tout_mult:
                logging.warn("* dead file %s duration: %d re-inserting",
                             file_num, duration)
                files_to_remove_from_files_running.append(file_num)
                self.files_available.append(file_num)

        for c in files_to_remove_from_files_running:
            del self.files_running[c]

    def find_something_to_solve(self):
        self.check_for_dead_files()
        logging.info("Num files_available pre-send: %d",
                     len(self.files_available))

        if len(self.files_available) == 0:
            return None

        file_num = self.files_available[0]
        del self.files_available[0]
        logging.info("Num files_available post-send: %d",
                     len(self.files_available))
        sys.stdout.flush()

        return file_num

    def handle_build(self, connection, cli_addr, indata):
        tosend = self.default_tosend()
        logging.info("Sending git revision %s to %s", options.git_rev,
                     cli_addr)
        send_command(connection, "build_data", tosend)

    def send_termination(self, connection, cli_addr):
        tosend = {}
        tosend["noshutdown"] = options.noshutdown
        send_command(connection, "finish", tosend)

        logging.info("No more to solve, terminating %s", cli_addr)
        global last_termination_sent
        last_termination_sent = time.time()

    def send_wait(self, connection, cli_addr):
        tosend = {}
        tosend["noshutdown"] = options.noshutdown
        logging.info("Everything is in sent queue, sending wait to %s", cli_addr)
        send_command(connection, "wait", tosend)

    def default_tosend(self):
        tosend = {}
        tosend["solver"] = options.solver
        tosend["git_rev"] = options.git_rev
        tosend["stats"] = options.stats
        tosend["gauss"] = options.gauss
        tosend["s3_bucket"] = options.s3_bucket
        tosend["given_folder"] = options.given_folder
        tosend["timeout_in_secs"] = options.timeout_in_secs
        tosend["mem_limit_in_mb"] = options.mem_limit_in_mb
        tosend["noshutdown"] = options.noshutdown
        tosend["extra_opts"] = options.extra_opts
        tosend["drat"] = options.drat
        tosend["region"] = options.region

        return tosend

    def send_one_to_solve(self, connection, cli_addr, file_num):
        # set timer that we have sent this to be solved
        self.files_running[file_num] = time.time()
        filename = self.files[file_num].name

        tosend = self.default_tosend()
        tosend["file_num"] = file_num
        tosend["cnf_filename"] = filename
        tosend["uniq_cnt"] = str(self.uniq_cnt)
        logging.info("Sending file %s (num %d) to %s",
                     filename, file_num, cli_addr)
        send_command(connection, "solve", tosend)
        self.uniq_cnt += 1

    def handle_need(self, connection, cli_addr, indata):
        # TODO don't ignore 'indata' for solving CNF instances, use it to
        # opitimize for uptime
        file_num = self.find_something_to_solve()

        if file_num is None:
            if len(self.files_running) == 0:
                self.send_termination(connection, cli_addr)
            else:
                self.send_wait(connection, cli_addr)
        else:
            self.send_one_to_solve(connection, cli_addr, file_num)

    def handle_one_client(self, conn, cli_addr):
        try:
            logging.info("connection from %s", cli_addr)

            data = get_n_bytes_from_connection(conn, 8)
            length = struct.unpack('!q', data)[0]
            data = get_n_bytes_from_connection(conn, length)
            data = pickle.loads(data)

            if data["command"] == "done":
                self.handle_done(conn, cli_addr, data)

            if data["command"] == "error":
                shutdown(-1)
                raise

            elif data["command"] == "need":
                self.handle_need(conn, cli_addr, data)

            elif data["command"] == "build":
                self.handle_build(conn, cli_addr, data)

            sys.stdout.flush()
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exc()
            the_trace = traceback.format_exc()

            logging.error("Exception from %s, Trace: %s", cli_addr,
                          the_trace)

        finally:
            # Clean up the connection
            logging.info("Finished with client %s", cli_addr)
            conn.close()

    def run(self):
        global acc_queue
        while True:
            conn, cli_addr = acc_queue.get()
            self.handle_one_client(conn, cli_addr)


class Listener (threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)

    def listen_to_connection(self):
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Bind the socket to the port
        server_address = ('0.0.0.0', options.port)
        logging.info('starting up on %s port %s', server_address, options.port)
        sock.bind(server_address)

        # Listen for incoming connections
        sock.listen(128)
        return sock

    def handle_one_connection(self):
        global acc_queue

        # Wait for a connection
        conn, client_addr = self.sock.accept()
        acc_queue.put_nowait((conn, client_addr))

    def run(self):
        try:
            self.sock = self.listen_to_connection()
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            the_trace = traceback.format_exc().rstrip().replace("\n", " || ")
            logging.error("Cannot listen on stocket! Traceback: %s", the_trace)
            shutdown(-1)
            raise
        while True:
            self.handle_one_connection()


class SpotManager (threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)
        self.spot_creator = RequestSpotClient.RequestSpotClient(
            options.git_rev,
            ("test" in options.cnf_list), noshutdown=options.noshutdown,
            count=options.client_count)

    def run(self):
        while True:
            try:
                if not server.ready_to_shutdown():
                    self.spot_creator.create_spots_if_needed()
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                the_trace = traceback.format_exc().rstrip().replace("\n", " || ")
                logging.error("Cannot create spots! Traceback: %s", the_trace)

            time.sleep(60)


def shutdown(exitval=0):
    toexec = "sudo shutdown -h now"
    logging.info("SHUTTING DOWN")

    # send email
    try:
        email_subject = "Server shutting down "
        if exitval == 0:
            email_subject += "OK"
        else:
            email_subject += "FAIL"

        full_s3_folder = get_s3_folder(
            options.given_folder,
            options.git_rev,
            options.solver,
            options.timeout_in_secs,
            options.mem_limit_in_mb)
        text = """Server finished. Please download the final data:

mkdir {0}
cd {0}
aws s3 cp --recursive s3://{1}/{0}/ .

Don't forget to:

* check volume
* check EC2 still running

So long and thanks for all the fish!
""".format(full_s3_folder, options.s3_bucket)
        send_email(email_subject, text, options.logfile_name)
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        the_trace = traceback.format_exc().rstrip().replace("\n", " || ")
        logging.error("Cannot send email! Traceback: %s", the_trace)

    if not options.noshutdown:
        os.system(toexec)

    exit(exitval)


def set_up_logging():
    form = '[ %(asctime)-15s  %(levelname)s  %(message)s ]'
    logformatter = logging.Formatter(form)

    try:
        os.unlink(options.logfile_name)
    except:
        pass
    fileHandler = logging.FileHandler(options.logfile_name)
    fileHandler.setFormatter(logformatter)
    logging.getLogger().addHandler(fileHandler)
    logging.getLogger().setLevel(logging.INFO)

if __name__ == "__main__":
    global options
    global args
    options, args = server_option_parser.parse_arguments()
    if options.drat:
        assert "cryptominisat" in options.solver

    global acc_queue
    acc_queue = Queue.Queue()
    last_termination_sent = None

    set_up_logging()
    logging.info("Server called with parameters: %s",
                 pprint.pformat(options, indent=4).replace("\n", " || "))

    if not options.git_rev:
        options.git_rev = get_revision(options.base_dir + options.solver, options.base_dir)
        logging.info("Revision not given, taking HEAD: %s", options.git_rev)

    server = Server()
    listener = Listener()
    spotmanager = SpotManager()
    listener.setDaemon(True)
    server.setDaemon(True)
    spotmanager.setDaemon(True)

    listener.start()
    server.start()
    time.sleep(20)
    spotmanager.start()

    while threading.active_count() > 0:
        time.sleep(0.5)
        if last_termination_sent is not None and server.ready_to_shutdown():
            diff = time.time() - last_termination_sent
            limit = 100
            if diff > limit:
                break

    shutdown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sys

if len(sys.argv) == 1:
    print("""Usage example:
./{name} satcomp14_updated satcomp16_updated > satcomp1416_updated
""".format(name=sys.argv[0]))
    exit(-1)

files = {}
for a in range(len(sys.argv)-1):
    with open(sys.argv[a+1], "r") as f:
        for l in f:
            l = l.strip()
            fname = l[l.find("/")+1:]
            files[fname] = l

for a,b in files.items():
    print(b)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# pylint: disable=invalid-name,line-too-long,too-many-locals,consider-using-sys-exit

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import sklearn
import sklearn.metrics
import re
import ast
import math
import time
import os.path
import sqlite3
import functools
try:
    from termcolor import cprint
except ImportError:
    termcolor_avail = False
else:
    termcolor_avail = True

from pprint import pprint
try:
    import mlflow
except ImportError:
    mlflow_avail = False
else:
    mlflow_avail = True


class QueryHelper:
    def __init__(self, dbfname):
        if not os.path.isfile(dbfname):
            print("ERROR: Database file '%s' does not exist" % dbfname)
            exit(-1)

        self.conn = sqlite3.connect(dbfname)
        self.c = self.conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.commit()
        self.conn.close()


class QueryFill (QueryHelper):
    def create_indexes(self, verbose=False, used_clauses="used_clauses"):
        t = time.time()
        print("Recreating indexes...")
        queries = """
        create index `idxclid6-4` on `reduceDB` (`clauseID`, `conflicts`)
        create index `idxclidUCLS-2` on `{used_clauses}` ( `clauseID`, `used_at`);
        create index `idxcl_last_in_solver-1` on `cl_last_in_solver` ( `clauseID`, `conflicts`);
        """.format(used_clauses=used_clauses)
        for l in queries.split('\n'):
            t2 = time.time()

            if verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))

    def delete_and_create_all(self):
        tables = ["short", "long", "forever"]

        t = time.time()
        q = """DROP TABLE IF EXISTS `used_later`;"""
        self.c.execute(q)

        for table in tables:
            q = """
                DROP TABLE IF EXISTS `used_later_{name}`;
            """
            self.c.execute(q.format(name=table))

        # Create and fill used_later_X tables
        q_create = """
        create table `used_later_{name}` (
            `clauseID` bigint(20) NOT NULL,
            `rdb0conflicts` bigint(20) NOT NULL,
            `used_later_{name}` bigint(20),
            `offset` bigint(20) NOT NULL
        );"""
        self.c.execute(q_create.format(name="short"))
        self.c.execute(q_create.format(name="long"))
        self.c.execute(q_create.format(name="forever"))

        # Create used_later table
        t = time.time()
        q = """
        create table `used_later` (
            `clauseID` bigint(20) NOT NULL,
            `rdb0conflicts` bigint(20) NOT NULL,
            `used_later` bigint(20)
        );"""
        self.c.execute(q)

        idxs = """
        create index `used_later_{name}_idx3` on `used_later_{name}` (used_later_{name});
        create index `used_later_{name}_idx1` on `used_later_{name}` (`clauseID`, rdb0conflicts, offset);
        create index `used_later_{name}_idx2` on `used_later_{name}` (`clauseID`, rdb0conflicts, used_later_{name}, offset);"""

        for table in tables:
            for l in idxs.format(name=table).split('\n'):
                self.c.execute(l)

        idxs = """
        create index `used_later_idx3` on `used_later` (`used_later`);
        create index `used_later_idx1` on `used_later` (`clauseID`, rdb0conflicts);
        create index `used_later_idx2` on `used_later` (`clauseID`, rdb0conflicts, used_later);
        """
        t = time.time()
        for l in idxs.split('\n'):
            self.c.execute(l)

        print("used_later* dropped and recreated T: %-3.2f s" % (time.time() - t))

    def fill_used_later(self, used_clauses="used_clauses"):
        # Fill used_later table
        t = time.time()
        q = """insert into used_later
        (
        `clauseID`,
        `rdb0conflicts`,
        `used_later`
        )
        SELECT
        rdb0.clauseID
        , rdb0.conflicts
        , count(ucl.used_at) as `useful_later`
        FROM
        reduceDB as rdb0
        left join {used_clauses} as ucl

        -- for any point later than now
        -- reduceDB is always present, used_later may not be, hence left join
        on (ucl.clauseID = rdb0.clauseID
            and ucl.used_at > rdb0.conflicts)

        join cl_last_in_solver
        on cl_last_in_solver.clauseID = rdb0.clauseID

        WHERE
        rdb0.clauseID != 0
        and cl_last_in_solver.conflicts >= rdb0.conflicts

        group by rdb0.conflicts, rdb0.clauseID;"""

        self.c.execute(q.format(used_clauses=used_clauses))
        print("used_later filled T: %-3.2f s" % (time.time() - t))

    def fill_used_later_X(self, name, duration, offset=0,
                          used_clauses="used_clauses",
                          forever=False):
        q_fill = """
        insert into used_later_{name}
        (
        `clauseID`,
        `rdb0conflicts`,
        `used_later_{name}`,
        `offset`
        )
        SELECT
        rdb0.clauseID
        , rdb0.conflicts
        , count(ucl.used_at) as `used_later_{name}`
        , {offset}

        FROM
        reduceDB as rdb0
        left join {used_clauses} as ucl

        -- reduceDB is always present, {used_clauses} may not be, hence left join
        on (ucl.clauseID = rdb0.clauseID
            and ucl.used_at > (rdb0.conflicts+{offset})
            and ucl.used_at <= (rdb0.conflicts+{duration}+{offset}))

        join cl_last_in_solver
        on cl_last_in_solver.clauseID = rdb0.clauseID

        WHERE
        rdb0.clauseID != 0
        and (cl_last_in_solver.conflicts >= (rdb0.conflicts + {duration} + {offset})
        or 1=={forever})

        group by rdb0.clauseID, rdb0.conflicts;"""

        t = time.time()
        self.c.execute(q_fill.format(
            name=name, used_clauses=used_clauses,
            duration=duration, offset=offset, forever=int(forever)))
        print("used_later_%s filled T: %-3.2f s" %
              (name, time.time() - t))


def write_mit_header(f):
    f.write("""/******************************************
Copyright (c) 2018, Mate Soos

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
***********************************************/\n\n""")

def parse_configs(confs):
    match = re.match(r"^([0-9]*)-([0-9]*)$", confs)
    if not match:
        print("ERROR: we cannot parse your config options: '%s'" % confs)
        exit(-1)

    conf_from = int(match.group(1))
    conf_to = int(match.group(2))+1
    if conf_to <= conf_from:
        print("ERROR: Conf range is not increasing")
        exit(-1)

    print("Running configs:", range(conf_from, conf_to))
    return conf_from, conf_to


def get_features(fname):
    best_features = []
    check_file_exists(fname)
    with open(fname, "r") as f:
        for l in f:
            l = l.strip()
            if l != "":
                best_features.append(l)

    return best_features


def helper_divide(dividend, divisor, df, features, verb, name=None):
    """
    to be used like:
    import functools
    divide = functools.partial(helper.divide, df=df, features=features, verb=options.verbose)
    """

    # dividend feature not present
    #if dividend not in features:
        #return None

    # divisorfeature not present
    #if divisor not in features:
        #return None

    # divide
    if verb:
        print("Dividing. dividend: '%s' divisor: '%s' " % (dividend, divisor))

    if name is None:
        name = "(" + dividend + "/" + divisor + ")"

    df[name] = df[dividend].div(df[divisor]+0.000000001)
    return name

def helper_larger_than(lhs, rhs, df, features, verb):
    """
    to be used like:
    import functools
    larger_than = functools.partial(helper.larger_than, df=df, features=features, verb=options.verbose)
    """

    # divide
    if verb:
        print("Calulating '%s' >: '%s' " % (lhs, rhs))

    name = "(" + lhs + ">" + rhs + ")"
    df[name] = (df[lhs] > df[rhs]).astype(int)
    return name

def helper_add(toadd, df, features, verb):
    """
    to be used like:
    import functools
    larger_than = functools.partial(helper.larger_than, df=df, features=features, verb=options.verbose)
    """

    # add
    if verb:
        print("Calulating: the feature addition of: %s", toadd)

    name = "("
    for i in range(1, len(toadd)):
        name = toadd[i]
        if i < len(toadd)-1:
            name+="+"
    name += ")"

    df[name] = df[toadd[0]]
    for i in range(1, len(toadd)):
        df[name] += df[toadd[i]]
    return name


def dangerous(conn):
    conn.execute("PRAGMA journal_mode = MEMORY")
    conn.execute("PRAGMA synchronous = OFF")


def drop_idxs(conn):
    q = """
    SELECT name FROM sqlite_master WHERE type == 'index'
    """
    conn.execute(q)
    rows = conn.fetchall()
    queries = ""
    for row in rows:
        print("Will delete index:", row[0])
        queries += "drop index if exists `%s`;\n" % row[0]

    t = time.time()
    for q in queries.split("\n"):
        conn.execute(q)

    print("Removed indexes: T: %-3.2f s"% (time.time() - t))


def get_columns(tablename, verbose, conn):
    q = "pragma table_info(%s);" % tablename
    conn.execute(q)
    rows = conn.fetchall()
    columns = []
    for row in rows:
        if verbose:
            print("Using column in table {tablename}: {col}".format(
                tablename=tablename, col=row[1]))
        columns.append(row[1])

    return columns

def query_fragment(tablename, not_cols, short_name, verbose, conn):
    cols = get_columns(tablename, verbose, conn)
    filtered_cols = list(set(cols).difference(not_cols))
    ret = ""
    for col in filtered_cols:
        ret += ", {short_name}.`{col}` as `{short_name}.{col}`\n".format(
            col=col, short_name=short_name)

    if verbose:
        print("query for short name {short_name}: {ret}".format(
            short_name=short_name, ret=ret))

    return ret


def not_inside(not_these, inside_here):
    for not_this in not_these:
        if not_this in inside_here:
            return False

    return True

def add_computed_szfeat_for_clustering(df):
    print("Adding computed clustering features...")

    todiv = []
    for x in list(df):
        not_these = ["std", "min", "mean", "_per_", "binary", "vcg", "pnr", "horn", "max"]
        if "szfeat_cur" in x and x[-3:] != "var" and not_inside(not_these, x):
            todiv.append(x)

    # relative data
    cols = list(df)
    for col in cols:
        if "szfeat_cur" in col:
            for divper in todiv:
                df["("+col+"/"+divper+")"] = df[col]/df[divper]
                df["("+col+"<"+divper+")"] = (df[col] < df[divper]).astype(int)

    print("Added computed features.")


# to check for too large or NaN values:
def check_too_large_or_nan_values(df, features):
    print("Checking for too large or NaN values...")
    # features = df.columns.values.flatten().tolist()
    index = 0
    for index, row in df[features].iterrows():
        for x, name in zip(row, features):
            try:
                np.isfinite(x)
            except:
                print("Name:", name)
                print("Prolbem with value:", x)
                print(row)

            if not np.isfinite(x) or x > np.finfo(np.float32).max:
                print("issue with data for features: ", name, x)
            index += 1

    print("Checking finished.")


def print_confusion_matrix(cm,
                           normalize=False,
                           title='Confusion matrix',
                           cmap=plt.cm.Blues):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    print(title)
    if mlflow_avail:
        mlflow.log_metric(title, cm[0][0])
    np.set_printoptions(precision=2)
    print(cm)


def calc_min_split_point(df, min_samples_split):
    split_point = int(float(df.shape[0])*min_samples_split)
    if split_point < 10:
        split_point = 10
    print("Minimum split point: ", split_point)
    return split_point


def calc_regression_error(data, features, to_predict, clf, toprint,
                  average="binary", highlight=False):
    X_data = data[features]
    y_data = data[to_predict]
    print("Number of elements:", X_data.shape)
    if data.shape[0] <= 1:
        print("Cannot calculate regression error, too few elements")
        return None
    y_pred = clf.predict(X_data)
    main_error = sklearn.metrics.mean_squared_error(y_data, y_pred)
    print("Mean squared error is: ", main_error)

    for start,end in [(0,10), (1,10), (10, 100), (100, 1000), (1000,10000), (10000, 1000000)]:
        x = "--> Strata  %10d <= %20s < %10d " % (start, to_predict, end)
        myfilt = data[(data[to_predict] >= start) & (data[to_predict] < end)]
        X_data = myfilt[features]
        y_data = myfilt[to_predict]
        y = " -- elements: {:20}".format(str(X_data.shape))
        if myfilt.shape[0] <= 1:
            #print("Cannot calculate regression error, too few elements")
            error = -1
        else:
            y_pred = clf.predict(X_data)
            error = sklearn.metrics.mean_squared_error(y_data, y_pred)
        print("%s %s msqe: %13.1lf" % (x, y, error))

    for start,end in [(0,3), (3,8), (8, 15), (15, 25), (25,50), (50, 100), (100, 1000000)]:
        x = "--> Strata  %10d <= %20s < %10d " % (start, "rdb0.glue", end)
        myfilt = data[(data["rdb0.glue"] >= start) & (data["rdb0.glue"] < end)]
        X_data = myfilt[features]
        y_data = myfilt[to_predict]
        y = " -- elements: {:20}".format(str(X_data.shape))
        if myfilt.shape[0] <= 1:
            error = -1
        else:
            y_pred = clf.predict(X_data)
            error = sklearn.metrics.mean_squared_error(y_data, y_pred)
        print("%s %s msqe: %13.1lf" % (x, y, error))

    return main_error


def conf_matrixes(data, features, to_predict, clf, toprint,
                  average="binary", highlight=False):
    # get data
    X_data = data[features]
    y_data = data[to_predict]
    print("Number of elements:", X_data.shape)
    if data.shape[0] <= 1:
        print("Cannot calculate confusion matrix, too few elements")
        return None, None, None, None

    # Preform prediction
    def f(x):
        if x > 0.5:
            return 1
        else:
            return 0

    y_pred = clf.predict(X_data)
    #print("type(y_pred[0]): ",  type(y_pred[0]))
    if type(y_pred[0]) == np.float32:
        y_pred = np.array([f(x) for x in y_pred])

    # calc acc, precision, recall
    accuracy = sklearn.metrics.accuracy_score(
        y_data, y_pred)

    precision = sklearn.metrics.precision_score(
        y_data, y_pred, pos_label=1, average=average)


    recall = sklearn.metrics.recall_score(
        y_data, y_pred, pos_label=1, average=average)

    # ROC AUC
    predsi = np.array(y_pred)
    y_testi = pd.DataFrame(y_data)["x.class"].squeeze()
    try:
        roc_auc = sklearn.metrics.roc_auc_score(y_testi, predsi)
    except:
        print("NOTE: ROC AUC is set to 0 because of completely one-sided OK/BAD")
        roc_auc = 0

    # record to mlflow
    if mlflow_avail:
        mlflow.log_metric(toprint + " -- accuracy", accuracy)
        mlflow.log_metric(toprint + " -- precision", precision)
        mlflow.log_metric(toprint + " -- recall", recall)
        mlflow.log_metric(toprint + " -- roc_auc", roc_auc)

    color = "white"
    bckgrnd = "on_grey"
    if highlight:
        color="green"
        bckgrnd = "on_grey"

    txt = "%s prec : %-3.4f  recall: %-3.4f accuracy: %-3.4f roc_auc: %-3.4f"
    vals = (toprint, precision, recall, accuracy, roc_auc)
    if termcolor_avail:
        cprint(txt % vals , color, bckgrnd)
    else:
        cprint(txt % vals)

    # Plot confusion matrix
    cnf_matrix = sklearn.metrics.confusion_matrix(
        y_true=y_data, y_pred=y_pred)
    print_confusion_matrix(
        cnf_matrix,
        title='Confusion matrix without normalization -- %s' % toprint)
    print_confusion_matrix(
        cnf_matrix, normalize=True,
        title='Normalized confusion matrix -- %s' % toprint)

    return roc_auc


def check_file_exists(fname):
    try:
        f = open(fname)
    except IOError:
        print("File '%s' not accessible" % fname)
        exit(-1)
    finally:
        f.close()


def clear_data_from_str(df):
    values2nums = {'luby': 0, 'glue': 1, 'geom': 2}
    df.loc[:, ('cl.cur_restart_type')] = \
        df.loc[:, ('cl.cur_restart_type')].map(values2nums)

    df.loc[:, ('rdb0.cur_restart_type')] = \
        df.loc[:, ('rdb0.cur_restart_type')].map(values2nums)

    df.loc[:, ('rst_cur.restart_type')] = \
        df.loc[:, ('rst_cur.restart_type')].map(values2nums)

    if "rdb1.cur_restart_type" in df:
        df.loc[:, ('rdb1.cur_restart_type')] = \
            df.loc[:, ('rdb1.cur_restart_type')].map(values2nums)


def output_to_classical_dot(clf, features, fname):
    sklearn.tree.export_graphviz(clf, out_file=fname,
                                 feature_names=features,
                                 class_names=clf.classes_,
                                 filled=True, rounded=True,
                                 special_characters=True,
                                 proportion=True)
    print("Run dot:")
    print("dot -Tpng {fname} -o {fname}.png".format(fname=fname))
    print("gwenview {fname}.png".format(fname=fname))


def print_feature_ranking(clf, X_train, top_num_features, features, plot=False):
    best_features = []
    importances = clf.feature_importances_
    std = np.std(
        [tree.feature_importances_ for tree in clf.estimators_], axis=0)
    indices = np.argsort(importances)[::-1]
    indices = indices[:top_num_features]
    myrange = min(X_train.shape[1], top_num_features)

    # Print the feature ranking
    print("Feature ranking:")

    for f in range(myrange):
        print("%-3d  %-55s -- %8.4f" %
              (f + 1, features[indices[f]], importances[indices[f]]))
        best_features.append(features[indices[f]])

    # Plot the feature importances of the clf
    if plot:
        plot_feature_importances(importances, indices, myrange, std, features)

    return best_features


def plot_feature_importances(importances, indices, myrange, std, features):
        plt.figure()
        plt.title("Feature importances")
        plt.bar(range(myrange), importances[indices],
                color="r", align="center",
                yerr=std[indices])
        plt.xticks(range(myrange), [features[x]
                                    for x in indices], rotation=45)
        plt.xlim([-1, myrange])


def cldata_add_computed_features(df, verbose):
    print("Adding computed features...")
    del df["cl.conflicts"]
    divide = functools.partial(helper_divide, df=df, features=list(df), verb=verbose)
    larger_than = functools.partial(helper_larger_than, df=df, features=list(df), verb=verbose)
    add = functools.partial(helper_add, df=df, features=list(df), verb=verbose)

    # relative overlaps
    print("Relative overlaps...")
    divide("cl.num_total_lits_antecedents", "cl.antec_sum_size_hist")

    # deleting this feature which is NONE
    del df["cl.antecedents_glue_long_reds_avg"]
    del df["cl.antecedents_glue_long_reds_max"]
    del df["cl.antecedents_glue_long_reds_min"]
    del df["cl.antecedents_glue_long_reds_var"]
    del df["cl.antecedents_long_red_age_avg"]
    del df["cl.antecedents_long_red_age_var"]
    del df["cl.decision_level_hist"]
    del df["sum_cl_use.first_confl_used"]
    del df["sum_cl_use.last_confl_used"]

    # ************
    # TODO decision level and branch depth are the same, right???
    # ************
    print("size/glue/trail rel...")
    divide("cl.trail_depth_level", "cl.trail_depth_level_hist")

    rst_cur_all_props = add(["rst_cur.propBinRed",
                            "rst_cur.propBinIrred",
                            "rst_cur.propLongRed",
                            "rst_cur.propLongIrred"])

    divide("cl.num_total_lits_antecedents", "cl.num_antecedents")

    # sum RDB
    orig_cols = list(df)
    for col in orig_cols:
        if ("rdb0" in col) and "restart_type" not in col:
            col2 = col.replace("rdb0", "rdb1")
            cboth = "("+col+"+"+col2+")"
            df[cboth] = df[col]+df[col2]

    rdb0_act_ranking_rel = divide("rdb0.act_ranking", "rdb0.tot_cls_in_db", name="rdb0_act_ranking_rel")
    rdb1_act_ranking_rel = divide("rdb1.act_ranking", "rdb1.tot_cls_in_db", name="rdb1_act_ranking_rel")
    rdb0_plus_rdb1_ranking_rel = add([rdb0_act_ranking_rel, rdb1_act_ranking_rel])

    divide("rdb0.sum_uip1_used", "cl.time_inside_solver")
    divide("rdb1.sum_uip1_used", "cl.time_inside_solver")
    divide("rdb0.sum_propagations_made", "cl.time_inside_solver")
    divide("rdb1.sum_propagations_made", "cl.time_inside_solver")

    divisors = [
        "cl.size_hist"
        , "cl.glue_hist"
        , "rdb0.glue"
        , "cl.orig_glue"
        , "cl.glue_before_minim"
        , "cl.glue_hist_queue"
        , "cl.glue_hist_long"
        # , "cl.decision_level_hist"
        , "cl.num_resolutions_hist_lt"
        # , "cl.trail_depth_level_hist"
        # , "cl.backtrack_level_hist"
        , "cl.branch_depth_hist_queue"
        , "cl.antec_overlap_hist"
        , "(cl.num_total_lits_antecedents/cl.num_antecedents)"
        , "cl.num_antecedents"
        , rdb0_act_ranking_rel
        , rdb1_act_ranking_rel
        #, "szfeat_cur.var_cl_ratio"
        , "cl.time_inside_solver"
        #, "((double)(rdb0.act_ranking_rel+rdb1.act_ranking_rel)/2.0)"
        #, "sqrt(rdb0.act_ranking_rel)"
        #, "sqrt(rdb1.act_ranking_rel)"
        #, "sqrt(rdb0_and_rdb1.act_ranking_rel_avg)"
        # , "cl.num_overlap_literals"
        # , "rst_cur.resolutions"
        #, "rdb0.act_ranking_top_10"
        ]

    # Thanks to Chai Kian Ming Adam for the idea of using LOG instead of SQRT
    # add LOG
    if True:
        toadd = []
        for divisor in divisors:
            x = "log2("+divisor+")"
            df[x] = df[divisor].apply(np.log2)
            toadd.append(x)
        divisors.extend(toadd)

    # relative data
    cols = list(df)
    for col in cols:
        if ("rdb" in col or "cl." in col or "rst" in col) and "restart_type" not in col and "tot_cls_in" not in col and "rst_cur" not in col:
            for divisor in divisors:
                divide(divisor, col)
                divide(col, divisor)

    divisors.extend([
        rst_cur_all_props
        , "rdb0.last_touched_diff"
        , "rdb0.used_for_uip_creation"
        , "rdb1.used_for_uip_creation"
        , "rdb0.propagations_made"
        , "rdb1.propagations_made"
        , "rdb0.sum_propagations_made"
        , "(rdb0.sum_propagations_made/cl.time_inside_solver)"
        , "(rdb1.sum_propagations_made/cl.time_inside_solver)"
        , "(rdb0.sum_uip1_used/cl.time_inside_solver)"
        , "(rdb1.sum_uip1_used/cl.time_inside_solver)"])

    # smaller/larger than
    if False:
        for col in cols:
            if ("rdb" in col or "cl." in col or "rst" in col) and "restart_type" not in col:
                for divisor in divisors:
                    larger_than(col, divisor)

    # satzilla stuff
    if False:
        divisors = [
            "szfeat_cur.numVars",
            "szfeat_cur.numClauses",
            "szfeat_cur.var_cl_ratio",
            "szfeat_cur.avg_confl_size",
            "szfeat_cur.avg_branch_depth",
            "szfeat_cur.red_glue_distr_mean"
        ]
        for col in orig_cols:
            if "szfeat" in col:
                for divisor in divisors:
                    if "min" not in divisor:
                        divide(col, divisor)
                        larger_than(col, divisor)

    # relative RDB
    if True:
        print("Relative RDB...")
        for col in orig_cols:
            if "rdb0" in col and "restart_type" not in col:
                rdb0 = col
                rdb1 = col.replace("rdb0", "rdb1")
                larger_than(rdb0, rdb1)

                raw_col = col.replace("rdb0.", "")
                if raw_col not in ["sum_propagations_made", "propagations_made", "dump_no", "conflicts_made", "used_for_uip_creation", "sum_uip1_used", "activity_rel", "last_touched_diff", "ttl"]:
                    print(rdb0)
                    divide(rdb0, rdb1)

    # smaller-or-greater comparisons
    print("smaller-or-greater comparisons...")
    larger_than("cl.antec_sum_size_hist", "cl.num_total_lits_antecedents")
    larger_than("cl.antec_overlap_hist", "cl.num_overlap_literals")

    # print("flatten/list...")
    #old = set(df.columns.values.flatten().tolist())
    #df = df.dropna(how="all")
    #new = set(df.columns.values.flatten().tolist())
    #if len(old - new) > 0:
        #print("ERROR: a NaN number turned up")
        #print("columns: ", (old - new))
        #assert(False)
        #exit(-1)



#!/usr/bin/env python

import sys
of = sys.argv[1]
f = open(of, "r")
text = f.read()
mylines = text.splitlines()
i = 0;
time = []
for line in mylines :
  time.append(float(line))
  #print "t:%f" %(time[i])
  i += 1

lastnum = -1
for a in range(0, 60000, 1) :
  num = 0
  for t in time :
    #print "t: %f a: %d" %(t, a)
    if (t < a) :
      num += 1
  
  if (lastnum != num) :
      print "%d \t%d" %(num, a)
  lastnum = num

f.close();



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2018  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import boto
import traceback
import sys
import subprocess
import socket
import fcntl
import struct
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
import smtplib
import ConfigParser
config = ConfigParser.ConfigParser()
config.read("/home/ubuntu/email.conf")


def send_email(subject, text, fname=None):
    msg = MIMEMultipart()
    msg['Subject'] = 'Email from solver: %s' % subject
    msg['From'] = 'msoos@msoos.org'
    msg['To'] = 'soos.mate@gmail.com'

    # That is what you see if you have no email client:
    msg.preamble = 'Multipart massage.\n'

    # Text part
    part = MIMEText(text)
    msg.attach(part)

    # Attachment(s)
    if fname:
        part = MIMEApplication(open(fname, "rb").read())
        part.add_header('Content-Disposition', 'attachment', filename="attachment.txt")
        msg.attach(part)

    # Connect to STMP server
    email_login = config.get("email", "login")
    email_pass = config.get("email", "pass")

    smtp = smtplib.SMTP_SSL("email-smtp.us-west-2.amazonaws.com")
    smtp.login(email_login, email_pass)

    # Send email
    smtp.sendmail(msg['From'], msg['To'], msg.as_string())


def get_ip_address(ifname):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        s.fileno(),
        0x8915,  # SIOCGIFADDR
        struct.pack('256s', ifname[:15])
    )[20:24])


def get_revision(full_solver_path, base_dir):
    revision = subprocess.check_output(['git', 'rev-parse', 'HEAD'])
    return revision.strip()


def get_s3_folder(folder, rev, solver, timeout, memout):
    print("folder: %s rev: %s tout: %s memout %s" % (folder, rev, timeout, memout))
    solver_exe = solver[solver.rfind("/")+1:]
    return folder + "-{rev}-{solver}-tout-{tout}-mout-{mout}".format(
        rev=rev[:9], solver=solver_exe, tout=timeout, mout=memout)
